-- --------------------------------------------------------------------------
-- Extra Farm
-- author: Blacky_BPG
-- creation date: 16.03.2014
-- 
-- Update to 1.0.1 24.03.2014 - fixed loading on empty silo
-- Update to 1.1.0 29.03.2014 - fixed pda functions, added support for AllInOneSilo by Marhu
-- --------------------------------------------------------------------------
-- Script for use multiple farms
-- 
-- --------------------------------------------------------------------------

MultiFarm = {};
MultiFarm_mt = Class(MultiFarm, Object);
InitObjectClass(MultiFarm, "MultiFarm");

function MultiFarm.onCreate(id)
	local object = MultiFarm:new(g_server ~= nil, g_client ~= nil);
	if object:load(id) then
		g_currentMission:addOnCreateLoadedObject(object);
		object:register(true);
	else
		object:delete();
	end;
end;

function MultiFarm:new(isServer, isClient, customMt)
	local mt = customMt;
	if mt == nil then
		mt = MultiFarm_mt;
	end;

	local self = Object:new(isServer, isClient, mt);
	self.fillTypes = {};
	self.MultiFarmDirtyFlag = self:getNextDirtyFlag();
	return self;
end;

function MultiFarm:delete()
	g_currentMission:removeOnCreateLoadedObjectToSave(self);
	if self.nodeId ~= 0 then
		g_currentMission:removeNodeObject(self.nodeId);
	end;
	MultiFarm:superClass().delete(self);
end;

function MultiFarm:getUnloadAmount(fillType, farmNumber)
	if fillType ~= nil and fillType ~= Fillable.FILLTYPE_UNKNOWN then
		if self.fillTypes[fillType].capacity > -1 then
			return (self.fillTypes[fillType].capacity - self.fillTypes[fillType].fillLevel);
		else
			return 1000000;
		end;
	end;
	return 0;
end;

function MultiFarm:getSiloAmount(fillType, farmNumber)
	if fillType ~= nil and fillType ~= Fillable.FILLTYPE_UNKNOWN then
		return math.max(0,self.fillTypes[fillType].fillLevel);
	end;
	return 0;
end;

function MultiFarm:setFillLevel(fillLevel, fillType, deltaFillLevel)
	local dirty = false;
	if fillType ~= nil and fillType ~= Fillable.FILLTYPE_UNKNOWN then
		if fillLevel ~= nil and fillLevel >= 0 then
			self.fillTypes[fillType].fillLevel = fillLevel;
			dirty = true;
		elseif deltaFillLevel ~= nil then
			self.fillTypes[fillType].fillLevel = self.fillTypes[fillType].fillLevel + deltaFillLevel;
			dirty = true;
		end;
		if dirty then
			for a, b in pairs(self.fillTypes[fillType].movingPlanes) do
				if b ~= nil then
					local amount = self.fillTypes[fillType].fillLevel;
					if amount < 0.001 then
						amount = 0;
					end;
					local x,y,z = getTranslation(b.nodeId);
					local y = b.moveMinY + (b.moveMaxY - b.moveMinY)*Utils.clamp(amount, 0, b.moveMaxAmount)/(b.moveMaxAmount);
					setTranslation(b.nodeId, x,y,z);
				end;
			end;
			for a, b in pairs(self.fillTypes[fillType].digitalDisplays) do
				if b ~= nil then
					local amount = self.fillTypes[fillType].fillLevel;
					if amount < 0.001 then
						amount = 0;
					end;
					b:setDisplay(amount);
				end;
			end;
			if self.isServer then
				self:raiseDirtyFlags(self.MultiFarmDirtyFlag);
			end;
		end;
	end;
end;

function MultiFarm:readStream(streamId, connection)
	MultiFarm:superClass().readStream(self, streamId, connection);
	if connection:getIsServer() then
		for i=1, Fillable.NUM_FILLTYPES do
			local index = streamReadInt32(streamId);
			local active = streamReadBool(streamId);
			local fillLevel = streamReadFloat32(streamId);
			self.fillTypes[index].active = active;
			self:setFillLevel(fillLevel,index);
		end;
	end;
end;

function MultiFarm:writeStream(streamId, connection)
	MultiFarm:superClass().writeStream(self, streamId, connection);
	if not connection:getIsServer() then
		for i=1, Fillable.NUM_FILLTYPES do
			streamWriteInt32(streamId, i);
			streamWriteBool(streamId, self.fillTypes[i].active);
			streamWriteFloat32(streamId, self.fillTypes[i].fillLevel);
		end;
	end;
end;

function MultiFarm:readUpdateStream(streamId, timestamp, connection)
	MultiFarm:superClass().readUpdateStream(self, streamId, timestamp, connection)
	if connection:getIsServer() then
		for i=1, Fillable.NUM_FILLTYPES do
			local index = streamReadInt32(streamId);
			local active = streamReadBool(streamId);
			local fillLevel = streamReadFloat32(streamId);
			self.fillTypes[index].active = active;
			self:setFillLevel(fillLevel,index);
		end;
	end;
end;

function MultiFarm:writeUpdateStream(streamId, connection, dirtyMask)
	MultiFarm:superClass().writeUpdateStream(self, streamId, connection, dirtyMask);
	if not connection:getIsServer() then
		for i=1, Fillable.NUM_FILLTYPES do
			streamWriteInt32(streamId, i);
			streamWriteBool(streamId, self.fillTypes[i].active);
			streamWriteFloat32(streamId, self.fillTypes[i].fillLevel);
		end;
	end;
end;

function MultiFarm:load(nodeId)
	self.nodeId = nodeId;

	local capacity = -1;
	local capacityStr = getUserAttribute(nodeId, "capacity");
	if capacityStr ~= nil then
		capacity = Utils.getNoNil(tonumber(capacityStr), self.capacity);
	end;

	self.tipTriggers = {};
	self.siloTriggers = {};
	self.movingPlanes = {};
	self.shovelFillTriggers = {};
	self.shovelTargets = {};
	self.digitalDisplays = {};
	self.gasStations = {};
	for i=1, Fillable.NUM_FILLTYPES do
		self.fillTypes[i] = {};
		self.fillTypes[i].index = i;
		self.fillTypes[i].active = false;
		self.fillTypes[i].fillLevel = 0;
		self.fillTypes[i].tipTriggers = {};
		self.fillTypes[i].siloTriggers = {};
		self.fillTypes[i].movingPlanes = {};
		self.fillTypes[i].digitalDisplays = {};
		self.fillTypes[i].shovelFillTriggers = {};
		self.fillTypes[i].shovelTargets = {};

		self.fillTypes[i].capacity = capacity;
	end;
	local gasStationsGroup = getUserAttribute(nodeId, "gasStationsGroup");
	if gasStationsGroup ~= nil then
		local gasStations = Utils.indexToObject(nodeId, gasStationsGroup);
		if gasStations ~= nil then
			local num = getNumOfChildren(gasStations);
			if num > 0 then
				for i=0, num-1 do
					local child = getChildAt(gasStations, i);
					if child ~= 0 then
						self.gasStations[child] = true;
					end;
				end;
			end;
		end;
	end;

	local tipTriggersGroup = getUserAttribute(nodeId, "tipTriggersGroup");
	if tipTriggersGroup ~= nil then
		local tipTriggers = Utils.indexToObject(nodeId, tipTriggersGroup);
		if tipTriggers ~= nil then
			local num = getNumOfChildren(tipTriggers);
			if num > 0 then
				for i=0, num-1 do
					local child = getChildAt(tipTriggers, i);
					if child ~= 0 then
						self.tipTriggers[child] = true;
					end;
				end;
			end;
		end;
	end;

	local siloTriggersGroup = getUserAttribute(nodeId, "siloTriggersGroup");
	if siloTriggersGroup ~= nil then
		local siloTriggers = Utils.indexToObject(nodeId, siloTriggersGroup);
		if siloTriggers ~= nil then
			local num = getNumOfChildren(siloTriggers);
			if num > 0 then
				for i=0, num-1 do
					local child = getChildAt(siloTriggers, i);
					if child ~= 0 then
						self.siloTriggers[child] = true;
					end;
				end;
			end;
		end;
	end;

	local shovelTargetsGroup = getUserAttribute(nodeId, "shovelTargetsGroup");
	if shovelTargetsGroup ~= nil then
		local shovelTargets = Utils.indexToObject(nodeId, shovelTargetsGroup);
		if shovelTargets ~= nil then
			local num = getNumOfChildren(shovelTargets);
			if num > 0 then
				for i=0, num-1 do
					local child = getChildAt(shovelTargets, i);
					if child ~= 0 then
						self.shovelTargets[child] = true;
					end;
				end;
			end;
		end;
	end;

	local shovelFillTriggersGroup = getUserAttribute(nodeId, "shovelFillTriggersGroup");
	if shovelFillTriggersGroup ~= nil then
		local shovelFillTriggers = Utils.indexToObject(nodeId, shovelFillTriggersGroup);
		if shovelFillTriggers ~= nil then
			local num = getNumOfChildren(shovelFillTriggers);
			if num > 0 then
				for i=0, num-1 do
					local child = getChildAt(shovelFillTriggers, i);
					if child ~= 0 then
						self.shovelFillTriggers[child] = true;
					end;
				end;
			end;
		end;
	end;

	local digitalDisplaysGroup = getUserAttribute(nodeId, "digitalDisplaysGroup");
	if digitalDisplaysGroup ~= nil then
		local digitalDisplays = Utils.indexToObject(nodeId, digitalDisplaysGroup);
		if digitalDisplays ~= nil then
			local num = getNumOfChildren(digitalDisplays);
			if num > 0 then
				for i=0, num-1 do
					local child = getChildAt(digitalDisplays, i);
					if child ~= 0 then
						self.digitalDisplays[child] = true;
					end;
				end;
			end;
		end;
	end;

	local movingPlanesGroup = getUserAttribute(nodeId, "movingPlanesGroup");
	if movingPlanesGroup ~= nil then
		local movingPlanes = Utils.indexToObject(nodeId, movingPlanesGroup);
		if movingPlanes ~= nil then
			local num = getNumOfChildren(movingPlanes);
			if num > 0 then
				for i=0, num-1 do
					local child = getChildAt(movingPlanes, i);
					if child ~= 0 then
						self.movingPlanes[child] = true;
						local num1 = getNumOfChildren(child);
						if num1 > 0 then
							for x=0, num1-1 do
								local child1 = getChildAt(child, x);
								if child1 ~= nil then
									self.shovelFillTriggers[child1] = true;
								end;
							end;
						end;
					end;
				end;
			end;
		end;
	end;

	self.checkForTriggers = 20;
	self.isEnabled = true;

	self.pagePDA = 0;
	self.pagePDAMin = 0;
	self.pagePDAMax = 0;

	g_currentMission:addNodeObject(self.nodeId, self);
	g_currentMission:addOnCreateLoadedObjectToSave(self);

	if g_currentMission.ExtraFarm == nil then g_currentMission.ExtraFarm = {}; end;
	self.farmNumber = table.getn(g_currentMission.ExtraFarm) + 1;
	g_currentMission.ExtraFarm[self.farmNumber] = self;

	local nameStr = getUserAttribute(nodeId, "farmName");
	if nameStr ~= nil then
		self.farmName = g_i18n:getText(nameStr);
	else
		self.farmName = g_i18n:getText("defaultFarm").." "..tostring(self.farmNumber);
	end;

	self.getSiloAmount = MultiFarm.getSiloAmount;
	self.getUnloadAmount = MultiFarm.getUnloadAmount;
	self.setFillLevel = MultiFarm.setFillLevel;
	return true;
end;

function MultiFarm:loadFromAttributesAndNodes(xmlFile, key)
	for i=1, Fillable.NUM_FILLTYPES do
		local fillKey = key..string.format(".fillType%d", i);
		if hasXMLProperty(xmlFile, fillKey) then
			local fillLevel = getXMLFloat(xmlFile, fillKey.."#fillLevel");
			if fillLevel ~= nil and fillLevel >= 0 then
				self:setFillLevel(fillLevel,i, nil);
			end;
		end;
	end;
	return true;
end;

function MultiFarm:getSaveAttributesAndNodes(nodeIdent)
	local attributes = "";
	local nodes = "";
	local c = 1;
	for i=1, Fillable.NUM_FILLTYPES do
		if self.fillTypes[i].active then
			if c > 1 then
				nodes = nodes.."\n";
			end;
			c = c + 1;
			nodes = nodes.. nodeIdent..'<fillType'..i..' fillLevel="'..self.fillTypes[i].fillLevel..'" />';
		end;
	end;
	return attributes, nodes;
end;

function MultiFarm:update(dt)
	if self.checkForTriggers > -1 then
		if self.checkForTriggers > 0 then
			self.checkForTriggers = self.checkForTriggers - 1;
		else
			self.checkForTriggers = -1;

			for k,v in pairs(g_currentMission.updateables) do
				if k ~= nil and v ~= nil then
					local shovelTrigger = nil;
					if v.nodeId ~= nil and self.shovelFillTriggers ~= nil and self.shovelFillTriggers[v.nodeId] ~= nil and self.shovelFillTriggers[v.nodeId] then
						shovelTrigger = v;
					end;
					if shovelTrigger ~= nil then
						shovelTrigger.extraSilo = self;
						table.insert(self.fillTypes[shovelTrigger.fillType].shovelFillTriggers,shovelTrigger);
						shovelTrigger.fillShovel = function(self, shovel, dt)
							local silo = self.extraSilo:getSiloAmount(self.fillType, self.extraSilo.farmNumber)
							if silo > 0 then
								local delta = shovel:fillShovelFromTrigger(self, silo, self.fillType, dt);
								if delta > 0 then
									self.extraSilo:setFillLevel(nil, self.fillType, -delta);
								end;
							end;
						end;
					end;
				end;
			end;

			for k,v in pairs(g_currentMission.nonUpdateables) do
				if k ~= nil and v ~= nil then
					local movingPlane = nil;
					local shovelTarget = nil;
					local digitalDisplay = nil;
					if v.nodeId ~= nil and v.isSilo ~= nil and self.shovelTargets ~= nil and self.shovelTargets[v.nodeId] ~= nil and self.shovelTargets[v.nodeId] then
						shovelTarget = v;
					end;
					if v.nodeId ~= nil and v.moveMaxAmount ~= nil and self.movingPlanes ~= nil and self.movingPlanes[v.nodeId] ~= nil and self.movingPlanes[v.nodeId] then
						movingPlane = v;
					end;
					if v.nodeId ~= nil and v.digitPosition ~= nil and self.digitalDisplays ~= nil and self.digitalDisplays[v.nodeId] ~= nil and self.digitalDisplays[v.nodeId] then
						digitalDisplay = v;
					end;
					if movingPlane ~= nil then
						movingPlane.extraSilo = self;
						g_currentMission:removeSiloAmountListener(movingPlane);
						movingPlane.onSiloAmountChanged = function(self, fillType, amount)
								return;
							end;
						table.insert(self.fillTypes[movingPlane.fillType].movingPlanes,movingPlane);
					end;
					if shovelTarget ~= nil then
						shovelTarget.extraSilo = self;
						for a=1, Fillable.NUM_FILLTYPES do
							if shovelTarget.fillTypes ~= nil and shovelTarget.fillTypes[a] ~= nil and shovelTarget.fillTypes[a] then
								table.insert(self.fillTypes[a].shovelTargets,shovelTarget);
								self.fillTypes[a].active = true;
								self.fillTypes[a].capacity = Utils.getNoNil(shovelTarget.capacity,self.fillTypes[a].capacity);
								shovelTarget.capacity = self.fillTypes[a].capacity;
							end;
						end;
						shovelTarget.addShovelFillLevel = function(self, shovel, fillLevelDelta, fillType)
							local allowed = self.extraSilo:getUnloadAmount(fillType, self.extraSilo.farmNumber);
							if self.fillTypes[fillType] ~= nil and allowed > 0 then
								if fillLevelDelta > allowed then
									fillLevelDelta = allowed;
								end;
								self.extraSilo:setFillLevel(nil, fillType, fillLevelDelta);
								return fillLevelDelta;
							end;
							return 0;
						end;
					end;
					if digitalDisplay ~= nil then
						digitalDisplay.extraSilo = self;
						digitalDisplay.maxSpace = self.fillTypes[digitalDisplay.fillType].capacity;
						g_currentMission:removeSiloAmountListener(digitalDisplay);
						digitalDisplay.onSiloAmountChanged = function(self, fillType, amount)
								return;
							end;
						table.insert(self.fillTypes[digitalDisplay.fillType].digitalDisplays,digitalDisplay);
					end;
				end;
			end;

			if g_currentMission.gasStations ~= nil then
				for k,v in pairs(g_currentMission.gasStations) do
					if k ~= nil and v ~= nil then
						local gasStation = nil;
						if v.rootNode ~= nil and v.trailerCan ~= nil and v.trailerOnly ~= nil and self.gasStations[v.rootNode] ~= nil and self.gasStations[v.rootNode] then
							gasStation = v;
						end;
						if gasStation ~= nil then
							gasStation.extraSilo = self;
						end;
					end;
				end;
			end;

			for k,v in pairs(g_currentMission.onCreateLoadedObjects) do
				if k ~= nil and v ~= nil then
					local tipTrigger = nil;
					local siloTrigger = nil;
					if v.rootNode ~= nil and v.triggerTipWidth ~= nil and self.tipTriggers ~= nil and self.tipTriggers[v.rootNode] ~= nil and self.tipTriggers[v.rootNode] then
						tipTrigger = v;
					end;
					if v.rootNode ~= nil and v.siloTrailerId ~= nil and self.siloTriggers ~= nil and self.siloTriggers[v.rootNode] ~= nil and self.siloTriggers[v.rootNode] then
						siloTrigger = v;
					end;
					-- Fix for AllIonOneSilo by Marhu
					if v.Trigger ~= nil and v.Trigger.rootNode ~= nil and self.siloTriggers ~= nil and self.siloTriggers[v.Trigger.rootNode] ~= nil and self.siloTriggers[v.Trigger.rootNode] then
						siloTrigger = v;
						self.siloTriggers[v.Trigger.rootNode] = nil;
					end;
					if tipTrigger ~= nil then
						tipTrigger.extraSilo = self;
						-- Compatibility for Extended Tip Trigger
						for a=1, Fillable.NUM_FILLTYPES do
							if tipTrigger.acceptedFillTypes ~= nil and tipTrigger.acceptedFillTypes[a] ~= nil and tipTrigger.acceptedFillTypes[a] then
								if tipTrigger.convertFruitsTo ~= nil then
									local fruit = Utils.getNoNil(tipTrigger.convertFruitsTo[a],a);
									table.insert(self.fillTypes[fruit].tipTriggers,tipTrigger);
									self.fillTypes[fruit].active = true;
									self.fillTypes[a].capacity = Utils.getNoNil(tipTrigger.capacity,self.fillTypes[a].capacity);
									if tipTrigger.maxAmountAllowed ~= nil and tipTrigger.maxAmountAllowed[a] ~= nil then
										self.fillTypes[a].capacity = tipTrigger.maxAmountAllowed[a];
									end;
								else
									table.insert(self.fillTypes[a].tipTriggers,tipTrigger);
									self.fillTypes[a].active = true;
									self.fillTypes[a].capacity = Utils.getNoNil(tipTrigger.capacity,self.fillTypes[a].capacity);
									if tipTrigger.maxAmountAllowed ~= nil and tipTrigger.maxAmountAllowed[a] ~= nil then
										self.fillTypes[a].capacity = tipTrigger.maxAmountAllowed[a];
									end;
								end;
							end;
						end;
						tipTrigger.updateTrailerTipping = function(self, trailer, fillDelta, fillType)
							if fillDelta < 0 then
								local cFruitsTo = fillType;
								if self.convertFruitsTo ~= nil then
									cFruitsTo = self.convertFruitsTo[fillType];
								end;
								if cFruitsTo ~= nil and cFruitsTo ~= fillType then
									fillType = cFruitsTo;
								end;
								fillDelta = -fillDelta;
								local oldFillDelta = fillDelta;
								if fillDelta > trailer.capacity/15 then
									fillDelta = math.min(trailer.capacity/15,444);
								end;
								local amount = self.extraSilo:getUnloadAmount(fillType, self.extraSilo.farmNumber);
								if amount < fillDelta then
									fillDelta = amount;
								end;
								if fillDelta > 0 then
									self.extraSilo:setFillLevel(nil, fillType, fillDelta);
								else
								end;
								if oldFillDelta ~= fillDelta then
									oldFillDelta = oldFillDelta - fillDelta;
									trailer:setFillLevel(trailer:getFillLevel(fillType) + oldFillDelta);
									if trailer:onEndTip() ~= nil then
										trailer:onEndTip()
									end;
								end;
								self:updateMoving(fillDelta);
							end;
						end;
						local tipTriggerDistance = tipTrigger.getTipDistanceFromTrailer
						tipTrigger.getTipDistanceFromTrailer = function(self, trailer, tipReferencePointIndex)
							if self.extraSilo:getUnloadAmount(trailer.currentFillType, self.extraSilo.farmNumber) <= 0 then
								return math.huge;
							end;
							return tipTriggerDistance(self, trailer, tipReferencePointIndex);
						end;
					end;

					if siloTrigger ~= nil then
						siloTrigger.extraSilo = self;
						-- Compatibility for SiloMultiTrigger and Fix for NonFruitTypes
						siloTrigger.mustPay = false;
						if siloTrigger.fillType ~= nil and siloTrigger.fillType == 0 then
							if siloTrigger.canLoad == nil and siloTrigger.fillTypePos == nil then
								local attrib = getUserAttribute(siloTrigger.rootNode, "fruitType");
								if attrib == nil then
									attrib = getUserAttribute(siloTrigger.rootNode, "fillType");
								end;
								if attrib ~= nil then
									local fT = Fillable.fillTypeNameToDesc[attrib];
									if fT ~= nil then
										siloTrigger.fillType = fT.index;
									end;
								end;
							end;
						end;
						if siloTrigger.canLoad ~= nil and siloTrigger.fillTypePos ~= nil then
							for a=1, Fillable.NUM_FILLTYPES do
								if siloTrigger.fillTypes ~= nil and siloTrigger.fillTypes[a] ~= nil then
									table.insert(self.fillTypes[a].siloTriggers,siloTrigger);
								end;
							end;
							siloTrigger.update = function(self,dt)
								if self.isServer then
									local trailer = self.siloTrailer;
									if trailer ~= nil and trailer:getAllowFillFromAir() then
										local smtMustStop = false;
										self.fillTypesPos = trailer.fillTypesPos;
										self.fillType = self.fillTypes[self.fillTypesPos];
										self.canLoad = trailer.canLoad;
										if self.fill >= 1 and not self.fillDone then
											if self.canLoad == true then
												trailer:resetFillLevelIfNeeded(self.fillType);
												local fillLevel = trailer:getFillLevel(self.fillType);
												local siloAmount = self.extraSilo:getSiloAmount(self.fillType, self.extraSilo.farmNumber);
												if siloAmount > 0 and trailer:allowFillType(self.fillType, false) then
													local deltaFillLevel = math.min(self.fillLitersPerSecond*0.001*dt, siloAmount);
													if trailer.smtLoadTo ~= nil then
														local maxFill = trailer.capacity * (trailer.smtLoadTo / 100);
														local smtTmp = trailer:getFillLevel(self.fillType);
														if smtTmp + deltaFillLevel > maxFill then
															deltaFillLevel = maxFill - smtTmp;
															smtMustStop = true;
														end;
													end;
													trailer:setFillLevel(fillLevel+deltaFillLevel, self.fillType);
													local newFillLevel = trailer:getFillLevel(self.fillType);
													if fillLevel ~= newFillLevel then
														fillLevel = newFillLevel - fillLevel;
														self.extraSilo:setFillLevel(nil, self.fillType, fillLevel);
														self:startFill();
													else
														self.fillDone = true;
														self:stopFill();
														self.canLoad = false;
													end;
												else
													self.fillDone = true;
													self:stopFill();
													self.canLoad = false;
												end;
											else
												self:stopFill();
											end;
										else
											self:stopFill();
										end;
										if smtMustStop then
											self:stopFill();
										end;
									else
										self:stopFill();
										self.canLoad = false;
										self.fillDone = true;
									end;
								end;
							end;
						elseif siloTrigger.Trigger ~= nil and siloTrigger.SiloBuyAuto ~= nil then
							for a=1, Fillable.NUM_FILLTYPES do
								if siloTrigger.FillTypes ~= nil and siloTrigger.FillTypes[a] ~= nil then
									table.insert(self.fillTypes[a].siloTriggers,siloTrigger);
								end;
							end;
							siloTrigger.OrgSiloBuyAuto = nil;
							siloTrigger.SiloBuyAuto = 1;
							siloTrigger.update = function(self, dt)
								local trailer = self.Trigger.siloTrailer;
								if g_currentMission:getIsClient() and self.Trigger.fill >= 4 and trailer ~= nil and trailer:getIsActiveForInput(false) then
									local SelectetFillType = self.SelectetFillType
									if InputBinding.hasEvent(InputBinding.ALLINONESILO_FRUCHT) then
										SelectetFillType = SelectetFillType + 1;
									end;
									self:SetFillType(SelectetFillType);
									local t = Fillable.fillTypeIndexToDesc[self.Trigger.fillType].nameI18N;
									local s = 0, "[" .. g_i18n:getText("fluid_unit_short") .. "]";
									local v = self.extraSilo:getSiloAmount(self.Trigger.fillType, self.extraSilo.farmNumber);
									g_currentMission:addHelpButtonText(string.format("%d%s %s", v, s, t), InputBinding.ALLINONESILO_FRUCHT);
									if self.IsFilling then
										g_currentMission:addHelpButtonText(g_i18n:getText("StopFill"), InputBinding.ALLINONESILO_START)
									else
										g_currentMission:addHelpButtonText(g_i18n:getText("StartFill"), InputBinding.ALLINONESILO_START)
									end;
									if InputBinding.hasEvent(InputBinding.ALLINONESILO_START) then
										self:SetIsFilling(not self.IsFilling);
									end;
								else
									self:SetIsFilling(false);
								end;
								if g_currentMission:getIsServer() then
									if self.Trigger.fill >= 4 and trailer ~= nil then
										if self.IsFilling and not self.Trigger.fillDone then
											local fillLevel = trailer:getFillLevel(self.Trigger.fillType);
											local siloAmount = self.extraSilo:getSiloAmount(self.Trigger.fillType, self.extraSilo.farmNumber);
											if siloAmount > 0 and trailer:allowFillType(self.Trigger.fillType, false) then
												local deltaFillLevel = math.min(self.Trigger.fillLitersPerSecond * 0.001 * dt, siloAmount);
												trailer:setFillLevel(fillLevel + deltaFillLevel, self.Trigger.fillType);
												local newFillLevel = trailer:getFillLevel(self.Trigger.fillType);
												if fillLevel ~= newFillLevel then
													self.extraSilo:setFillLevel(nil, self.Trigger.fillType, (fillLevel - newFillLevel));
													self.Trigger:startFill()
												end;
											else
												self:SetfillDone();
											end;
										end;
									else
										self:SetIsFilling(false);
									end;
								else
									self:SetIsFilling(false);
								end;
							end;
						else
							if siloTrigger.fillType ~= nil and siloTrigger.fillType > 0 then
								table.insert(self.fillTypes[siloTrigger.fillType].siloTriggers,siloTrigger);
								siloTrigger.update = function(self, dt)
									if self.isServer then
										local trailer = self.siloTrailer;
										if self.fill >= 4 and trailer ~= nil and not self.fillDone then
											trailer:resetFillLevelIfNeeded(self.fillType);
											local fillLevel = trailer:getFillLevel(self.fillType);
											local siloAmount = self.extraSilo:getSiloAmount(self.fillType, self.extraSilo.farmNumber);
											if siloAmount > 0 and trailer:allowFillType(self.fillType, false) then
												local deltaFillLevel = math.min(self.fillLitersPerSecond*0.001*dt, siloAmount);
												trailer:setFillLevel(fillLevel+deltaFillLevel, self.fillType);
												local newFillLevel = trailer:getFillLevel(self.fillType);
												if fillLevel ~= newFillLevel then
													self.extraSilo:setFillLevel(nil, self.fillType, (fillLevel - newFillLevel));
													self:startFill();
												else
													self.fillDone = true;
													self:stopFill();
												end;
											else
												self.fillDone = true;
												self:stopFill();
											end;
										end;
									end;
								end;
							end;
						end;
					end;
				end;
			end;
			local actives = 0;
			for i=1, Fillable.NUM_FILLTYPES do
				if self.fillTypes[i].active then
					local tippers = table.getn(self.fillTypes[i].tipTriggers) + table.getn(self.fillTypes[i].shovelTargets);
					local fillers = table.getn(self.fillTypes[i].siloTriggers) + table.getn(self.fillTypes[i].shovelFillTriggers);
					if tippers > 0 and fillers == 0 then
						-- Error weil TipTrigger aber keine SiloTrigger
						print("WARNING: No SiloTrigger or ShovelFillTrigger for Filltype "..tostring(Fillable.fillTypeIndexToDesc[i].name).." for MultiFarm "..tostring(self.farmNumber));
					elseif tippers == 0 and fillers > 0 then
						-- Error weil SiloTrigger aber keine TipTrigger
						print("WARNING: No TipTrigger or ShovelTargets for Filltype "..tostring(Fillable.fillTypeIndexToDesc[i].name).." for MultiFarm "..tostring(self.farmNumber));
					end;
					actives = actives + 1;
					self:setFillLevel(self.fillTypes[i].fillLevel,i);
				end;
			end;
			if actives > 0 then
				self.pagePDAMax = table.getn(g_currentMission.missionPDA.statsPages);
				local maxSelfPages = math.ceil(math.min(actives / 11,1));
				self.pagePDAMin = self.pagePDAMax + 1;
				for i=1, maxSelfPages do
					g_currentMission.missionPDA:addStatsPage(self.drawPDA, self, nil,self.pagePDAMax + i);
				end;
			end;
			self.tipTriggers = {};
			self.siloTriggers = {};
			self.movingPlanes = {};
			self.shovelFillTriggers = {};
			self.shovelTargets = {};
			self.digitalDisplays = {};
		end;
	end;
	if g_currentMission.missionPDA.pdaRowSpacing > 0.019 then
		  g_currentMission.missionPDA.pdaRowSpacing = g_currentMission.missionPDA.pdaRowSpacing * 0.9;
	end;
end;

function MultiFarm:drawPDA(a,b,c,d)
	local pda = g_currentMission.missionPDA;
	local curPage = Utils.getNoNil(pda.showStatsPage,0) - self.pagePDAMax;
	if curPage > 0 then
		if curPage == 1 then
			self.pagePDAlastFruit = 1;
		end;
		setTextColor(1, 1, 1, 1);
		setTextBold(true);
		local yOffset = 0.01;
		local i = 0.5;
		local idef = 1.0;
		renderText(pda.pdaCol1, pda.pdaHeadRow - pda.pdaRowSpacing * i + yOffset, pda.pdaFontSize+0.0019, self.farmName)
		i = 1.5;
		setTextBold(false);
		local maxType = (curPage * 11) + 1;
		local minType = maxType - 10;
		local lastFruit = self.pagePDAlastFruit;
		while minType < maxType do
			if lastFruit > Fillable.NUM_FILLTYPES then
				break;
			end;
			local name = nil;
			if self.fillTypes[lastFruit].active then
				minType = minType + 1;
				self.pagePDAlastFruit = math.max(lastFruit+1,64);
				local desc = Fillable.fillTypeIndexToDesc[lastFruit];
				name = "Fruit "..tostring(lastFruit);
				if desc ~= nil then
					name = g_i18n:getText(desc.name)
				end;
				renderText(pda.pdaCol1, pda.pdaHeadRow - pda.pdaRowSpacing * i + yOffset, pda.pdaFontSize, name .. " [" .. g_i18n:getText("fluid_unit_short") .. "]")
				renderText(pda.pdaCol2, pda.pdaHeadRow - pda.pdaRowSpacing * i + yOffset, pda.pdaFontSize, string.format("%1.0f", g_i18n:getFluid(self.fillTypes[lastFruit].fillLevel)))
				i = i + idef;
			end;
			lastFruit = lastFruit + 1;
		end;
	end;
	setTextColor(1, 1, 1, 1);
	setTextBold(false);
end;

function MultiFarm:updateTick(dt)
end;

g_onCreateUtil.addOnCreateFunction("MultiFarm", MultiFarm.onCreate);
